from dataclasses import dataclass
import typing as t

import energydatamodel as edm

@dataclass
class Dataset:
    name: str
    description: str
    energy_system: edm.EnergySystem
    data: t.Optional[edm.TimeSeries] = None
    
    pass


from dataclasses import dataclass, fields
import numpy as np

from emflow.spaces import BaseSpace


@dataclass
class OutputSpace(BaseSpace):
    """
    The output space for the energy model.
    """

ActionSpace = OutputSpace


from abc import ABC, abstractmethod

class Agent(ABC):
    def __init__(self):
        # Initialize agent properties
        pass
    
    @abstractmethod
    def choose_action(self, state):
        """
        Choose an action based on the given state.

        Parameters:
            state: The current state of the environment.

        Returns:
            action: The action chosen by the agent.
        """
        # Implement logic to choose an action
        # For example, a random choice:
        # return env.action_space.sample()
        pass

    def learn(self, state, action, reward, next_state):
        """
        Learn from the experience (used in training).

        Parameters:
            state: The current state of the environment.
            action: The action taken in the state.
            reward: The reward received from the environment.
            next_state: The next state of the environment.
        """
        # Implement learning process
        pass



import typing as t
import gymnasium as gym
from gymnasium import spaces
from dataclasses import dataclass, fields
import numpy as np

import pandas as pd
import energydatamodel as edm
from emflow.models import Agent, Optimizer, Predictor, Simulator


@dataclass
class BaseVector:
    """
    Base vector.
    """

    @property
    def vector(self) -> np.ndarray:
        """ 
        Get all fields (attributes) of the dataclass as a vector. 
        """

        all_fields = fields(self)        
        values = [getattr(self, field.name) for field in all_fields]
        array = np.array(values)
        return array

    @classmethod
    def from_vector(cls, state_vector):
        """ 
        Create the :class:BaseVector from a :class:numpy.array. 
        """

        attribute_names = [field.name for field in fields(cls)]

        if len(state_vector) != len(attribute_names):
            raise ValueError("State vector and attribute names must have the same length")

        attribute_values = dict(zip(attribute_names, state_vector))
        state = cls(**attribute_values)
        return state


@dataclass
class BaseState(BaseVector):
    pass


@dataclass
class BaseAction(BaseVector):
    pass


@dataclass
class BaseEnvironment(gym.Env):
    pass


@dataclass
class BaseObjective:
    pass


@dataclass
class BaseProblem:
    system: t.Union[edm.Site, edm.EnergySystem]
    state: BaseState
    action: BaseAction
    environment: BaseEnvironment
    objective: BaseObjective
    model: t.Union[Agent, Optimizer, Predictor, Simulator]





from .base import BaseVector, BaseState, BaseAction, BaseEnvironment, BaseObjective, BaseProblem

from .models.agent import Agent
from .models.optimizer import Optimizer
from .models.predictor import Predictor
from .models.simulator import Simulator
#from energydatamodel.pv import PVArray


from dataclasses import dataclass, fields
import numpy as np

from emflow.spaces import BaseSpace


@dataclass
class InputSpace(BaseSpace):
    """
    The input space for the energy model.
    """

StateSpace = InputSpace


from abc import ABC, abstractmethod

class Predictor(ABC):
    def __init__(self):
        """
        Initialize the predictor.
        """
        pass

    def load_data(self):
        """
        Load data used for training the predictor. 

        This is a recommended method, but not mandatory. Subclasses may override it.
        """
        # Implement data loading method here. 
        pass

    def create_features(self):
        """
        Create features from the loaded data used for training the predictor.
        Load data used for training sthe predictor. 

        This is a recommended method, but not mandatory. Subclasses may override it.
        """
        # Implement feature creation method here.
        pass

    def train(self):
        """
        Train the predictor from the training data.

        This is a recommended method, but not mandatory. Subclasses may override it.
        """
        # Implement model training method here.
        pass
    
    @abstractmethod
    def predict(self, input):
        """
        Make a prediction based on the input data.

        This is a mandatory method that must be implemented by subclasses

        Parameters:
            input: Data on which prediction is to be made.

        Returns:
            prediction: The output of the model.
        """
        # Implement prediction method here. 
        pass


import typing as t
from abc import ABC, abstractmethod

from emflow.spaces import StateSpace

class Simulator(ABC):
    def __init__(self, initial_state):
        """
        Initialize the Simulator with a given initial state.

        :param initial_state: The initial state of the simulator.
        """
        self.state = initial_state
        # Initialize other necessary properties

    @abstractmethod
    def _transition_function(self, state, action):
        """
        Define the state transition logic.

        :param state: The current state.
        :param action: The action taken.
        :return: The next state.
        """
        # Implement how the state changes in response to the action
        # This is where the core simulation logic is defined
        pass

    @abstractmethod
    def _gather_info(self):
        """
        Gather additional information about the current state, if necessary.

        :return: A dictionary containing additional info.
        """
        # This can include details like simulation metrics, debug info, etc.
        pass

    @abstractmethod
    def reset(self):
        """
        Reset the simulator to the initial state or a new start state.

        :return: The reset state.
        """
        # Reset the state to its initial configuration or a new state
        pass

    @abstractmethod
    def step(self, action=None) -> t.Tuple[StateSpace, dict]:
        """
        Update the simulator's state based on an action.

        :param action: The action to be applied to the state.
        :return: A tuple containing the next state and any additional info.
        """
        next_state = self._state_transition_function(self.state, action)
        self.state = next_state
        info = self._gather_info()

        return next_state, info





from dataclasses import dataclass, fields
import numpy as np

@dataclass
class BaseSpace:
    """
    A base space with methods for converting data formats.

    """

    @classmethod
    def from_array(cls, input_array: np.ndarray) -> "BaseSpace":
        """
        Create the action from a numpy array.
        """
        action = cls(*input_array)
        return action

    def to_array(self) -> np.ndarray:
        """
        Convert the action to a numpy array.
        """
        all_fields = fields(self)        
        values = [getattr(self, field.name) for field in all_fields]
        output_array = np.array(values)
        return output_array

    @property
    def vector(self) -> np.ndarray:
        return self.to_array()

    @classmethod
    def from_tuple(cls, input_tuple):
        """
        Create the action from a tuple.
        """        
        return cls(*input_tuple)

    def to_tuple(self):
        """
        Convert the action to a tuple.
        """       
        all_fields = fields(self)        
        values = [getattr(self, field.name) for field in all_fields]
        output_tuple = tuple(values)
        return output_tuple

    @property
    def tuple(self) -> tuple:
        return self.to_tuple()


"""This example showcases how to use emflow for the solar forecasting task of GEFCom2014."""
import sys, os
sys.path.insert(0, os.path.abspath('../../../'))

import emflow
import emflow as emf

import gym
from gym import spaces
from dataclasses import dataclass
import numpy as np

import energydatamodel as edm

# Problems: 
# 1) How do I define an initial state with a different size compared to the consecutive states? 
# 2) How do I provide datetime index information as part of the state? Should I provide a pandas as the state information instead of numpy? 

def define_system(): 
    timeseries_1 = edm.TimeSeries(df=df, column_name=('1', 'POWER'))
    pvsystem_1 = edm.PVArray(capacity=1,
                            longitude=145,
                            latitude=-37.5,
                            altitude=595,
                            surface_azimuth=38,
                            surface_tilt=36,
                            timeseries=timeseries_1)

    timeseries_2 = edm.TimeSeries(df=df, column_name=('2', 'POWER'))
    pvsystem_2 = edm.PVArray(capacity=1,
                            longitude=145,
                            latitude=-37.5,
                            altitude=602,
                            surface_azimuth=327,
                            surface_tilt=35,
                            timeseries=timeseries_2)

    timeseries_3 = edm.TimeSeries(df=df, column_name=('3', 'POWER'))
    pvsystem_3 = edm.PVArray(capacity=1,
                            longitude=145,
                            latitude=-37.5,
                            altitude=951,
                            surface_azimuth=31,
                            surface_tilt=21,
                            timeseries=timeseries_3)

    portfolio = edm.Portfolio(assets=[pvsystem_1, pvsystem_2, pvsystem_3])

    return portfolio


@dataclass
class State(emf.BaseState):
    """
    The state vector.

    Args:
        pv_production [kWh/h]: Power production from solar wind used in the microgrid.
    """

    pv_production_1: gym.spaces.Box
    pv_production_2: gym.spaces.Box
    pv_production_3: gym.spaces.Box


@dataclass
class Action:
    test: float = 4

class GEFCom2014SolarEnv(gym.Env):
    """An Environment for GEFCom2014 Solar Competition."""
    metadata = {'render.modes': ['human']}

    def __init__(self, portfolio, df):
        super(GEFCom2014SolarEnv, self).__init__()
        # Define action and observation space
        # They must be gym.spaces objects

        self.portfolio = portfolio
        self.df = df
        self.initial_state_space = self._define_initial_state_space()
        self.state_space = self._define_task_state_space()
        self.action_space = self._define_task_state_space()

    def step(self, action):
        # Execute one time step within the environment
        # return observation, reward, done, info
        return 0

    def reset(self):
        # Reset the state of the environment to an initial state
        self.state = self._set_initial_state()
        return self._get_observation()

    def render(self, mode='human', close=False):
        # Render the environment to the screen
        pass

    def close(self):
        # Clean up resources
        pass

    def _define_initial_state_space(self):
        initial_start = "2012-04-01 01:00:00"
        initial_end = "2013-04-01 00:00:00"
        initial_datapoints = len(self.df[initial_start:initial_end])
        initial_state = self._define_state_space(initial_datapoints)
        return initial_state

    def _define_task_state_space(self):
        task1_start = "2013-04-01 01:00:00"       
        task1_end = "2013-04-01 01:00:00"       
        task_datapoints = len(self.df[task1_start:task1_end])
        task_state = self._define_state_space(task_datapoints)
        return task_state

    def _define_state_space(self, datapoints):
        pv_production_1 = spaces.Box(low=0, high=self.portfolio.assets[0].capacity, shape=(datapoints, 1), dtype=np.float32)
        pv_production_2 = spaces.Box(low=0, high=self.portfolio.assets[1].capacity, shape=(datapoints, 1), dtype=np.float32)
        pv_production_3 = spaces.Box(low=0, high=self.portfolio.assets[2].capacity, shape=(datapoints, 1), dtype=np.float32)
        state = spaces.Tuple((pv_production_1, pv_production_2, pv_production_3))
        return state
    
    def _get_observation(self):
        return self.state


"""This module contains the MicroGrid environment for OpenAI Gym."""

import gym
from gym import spaces
from dataclasses import dataclass
import numpy as np

@dataclass
class State:
    """The state vector.

    Args:
        consumption [kWh/h]: Consumption in the microgrid.
        wind_production [kWh/h]: Power production from solar used in the microgrid.
        pv_production [kWh/h]: Power production from solar wind used in the microgrid.
        battery_storage [kWh]: Potential energy stored in the battery.
        hydrogen_storage [kWh]: Potential energy stored in the hydrogen.
        grid_import [kWh/h]: Power imported from the grid to the microgrid.
        grid_import_peak [kWh/h]: Peak power imported from the grid to the microgrid.
        spot_market_price [NOK/kWh]: The spot market price for Trondheim.
    """

    consumption: float
    pv_production: float
    wind_production: float
    battery_storage: float
    hydrogen_storage: float
    grid_import: float
    grid_import_peak: float
    spot_market_price: float

    @property
    def vector(self) -> np.ndarray:
        return np.array(
            [
                self.consumption,
                self.pv_production,
                self.wind_production,
                self.battery_storage,
                self.hydrogen_storage,
                self.grid_import,
                self.grid_import_peak,
                self.spot_market_price,
            ],
            dtype=np.float64,
        )

    @classmethod
    def from_vector(cls, state: np.ndarray) -> "State":
        return cls(
            consumption=state.item(0),
            pv_production=state.item(1),
            wind_production=state.item(2),
            battery_storage=state.item(3),
            hydrogen_storage=state.item(4),
            grid_import=state.item(5),
            grid_import_peak=state.item(6),
            spot_market_price=state.item(7),
        )


class MicroGridEnv(gym.Env):
    """Custom Environment for a Microgrid simulation"""
    metadata = {'render.modes': ['human']}

    def __init__(self):
        super(MicroGridEnv, self).__init__()
        # Define action and observation space
        # They must be gym.spaces objects
        # Example: self.action_space = spaces.Discrete(2)
        # Example: self.observation_space = spaces.Box(low=0, high=255, shape=(...), dtype=np.float32)

    def step(self, action):
        # Execute one time step within the environment
        # return observation, reward, done, info
        return 0

    def reset(self):
        # Reset the state of the environment to an initial state
        pass

    def render(self, mode='human', close=False):
        # Render the environment to the screen
        pass

    def close(self):
        # Clean up resources
        pass
# Test your environment
env = MicroGridEnv()



from abc import ABC, abstractmethod

class Optimizer(ABC):
    def __init__(self):
        """
        Initialize the Optimizer.
        """
        # Initialization code here

    @abstractmethod
    def optimize(self, objective_function, constraints=None):
        """
        Perform the optimization.

        :param objective_function: The objective function to be minimized or maximized.
        :param constraints: (Optional) Constraints for the optimization problem.
        :return: The result of the optimization.
        """
        # Implement optimization logic here
        pass



from setuptools import setup, find_packages

def read_requirements():
    with open('requirements.txt', 'r') as file:
        return file.read().splitlines()

setup(
    name='enerflow',
    version='0.0.1',
    packages=find_packages(),
    license='MIT',
    description='⚡ Open-source framework for sequential decision problems in the energy domain',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    author='rebase.energy',
    author_email='hello@rebase.energy',
    url='https://github.com/rebase-energy/enerflow',
    install_requires=read_requirements()
)


from .dataset import Dataset
from .environments import MicroGridEnv, HybridPowerPlantEnv, EnergyCommunityEnv
from .objectives import MeanSquaredError
from .checks import EnergyBalanceCheck, MassBalanceCheck
from .problem import Problem



from dataclasses import dataclass
import typing as t

import energydatamodel as edm
import emflow as ef

@dataclass
class Problem:
    """
    A problem to be solved by the energy model.
    """
    name: str
    description: str
    energy_system: edm.EnergySystem
    dataset: tuple

